/****************** AHB Interface PSL Sugar Assertions ****************/
`timescale 1 ns / 100 ps
module ahbCompliance (
//	  hAddr_i, hTrans_i, hWrite_i, hSize_i,
//	  hBurst_i, hProt_i, hWData_i, hRData_i, hReady_i, hResp_i,
//	  hReq_i, hLock_i, hGnt_i, hResetn_i, hSel_i, hMaster_i,
//	  hMastLock_i, hSplit_i, hClk_i);
  input hclk,         // Bus clock,  rising edge of HCLK.
  input hresetn,      // Reset  active LOW 
  input haddr[31:0],  // Address bus, Master 
  input htrans[1:0],  // Master Transfer type,  NONSEQUENTIAL, SEQUENTIAL, IDLE or BUSY.
  input hwrite,       // Transfer direction, HIGH =>  write; LOW => read 
  input hsize[2:0],   // Transfer size, byte (8-bit), halfword (16-bit) or word (32-bit).
  input hburst[2:0],  // Burst type, Four, eight and sixteen beat, incrementing or wrapping.
  input hprot[3:0],   // Protection control
  
  input hwdata[31:0], // master Write data bus
  input HSELx,        // Slave select
  input hrdata[31:0], // Slave Read data bus
  input hready,       // Slave Transfer done, HIGH => finished; LOW => extend transfer.
  input hresp[1:0]);  // Transfer response: OKAY, ERROR, RETRY and SPLIT.

  localparam dataBusWidth = 32;
  localparam numSlaves = 32;
  localparam busMaster = 0;
  /* these localparams define the Transfer Type as specified by HTrans */
  localparam IDLE = 2'b00;
  localparam BUSY = 2'b01;
  localparam SEQ = 2'b11;
  localparam NONSEQ = 2'b10;
  /* these localparams define the Burst Mode as specified by HBurst */
  localparam SINGLE = 3'b000;
  localparam INCR = 3'b001;
  localparam WRAP4 = 3'b010;
  localparam INCR4 = 3'b011;
  localparam WRAP8 = 3'b100;
  localparam INCR8 = 3'b101;
  localparam WRAP16 = 3'b110;
  localparam INCR16 = 3'b111;
  /* define the amount of data in each beat as specified by HSize*/
  localparam bits8 = 3'b000;
  localparam bits16 = 3'b001;
  localparam bits32 = 3'b010;
  localparam bits64 = 3'b011;
  localparam bits128 = 3'b100;
  localparam bits256 = 3'b101;
  localparam bits512 = 3'b110;
  localparam bits1024 = 3'b111;
  /* define the transfer response signals as specified by HResp*/
  localparam OKAY = 2'b00;
  localparam ERROR = 2'b01;
  localparam RETRY = 2'b10;
  localparam SPLIT = 2'b11;
  
  input [31:0] hAddr, hWData, hRData;
  input [1:0]  hTrans, hResp;
  input 	   hWrite, hReady, hReq;
  input [2:0]  hSize, hBurst;
  input [3:0]  hProt, hMaster;
  input 	   hLock, hGnt, hMastLock;
  input [numSlaves-1:0] hSel;
  input [15:0] 			hSplit;
  input 				hResetn, hClk;
  
  reg [31:0] 			prev_hAddr, prev_hAddrPlusSize;
  reg 					prev_hWrite, prev_hReady;
  reg [numSlaves-1:0] 	prev_hSel;
  reg [3:0] 			prev_hProt;
  reg [2:0] 			prev_hSize, prev_hBurst;
  reg [dataBusWidth-1:0] prev_hRData, prev_hWData;
  reg [1:0] 			 prev_hTrans, prev_hResp;
  integer 				 AddrIncrement;
  integer 				 NumberBeats=0;
  /* capture data that is needed in the following assertions */
  always @(posedge hClk or negedge hResetn)
  begin
	prev_hAddr <= hAddr;
	prev_hReady <= hReady;
	prev_hWrite <= hWrite;
	prev_hRData <= hRData;
	prev_hWData <= hWData;
	prev_hTrans <= hTrans;
	prev_hBurst <= hBurst;
	prev_hResp <= hResp;
	prev_hSel <= hSel;
	prev_hSize <= hSize;
	prev_hProt <= hProt;
	case (hSize) /* AddrIncrement is used to check the address */
	  bits8: AddrIncrement=1;
	  bits16: AddrIncrement=2;
	  bits32: AddrIncrement=4;
	  bits64: AddrIncrement=8;
	  bits128: AddrIncrement=16;
	  bits256: AddrIncrement=32;
	  bits512: AddrIncrement=64;
	  bits1024: AddrIncrement=128;
	  default: ;
	endcase
	/* NumberBeats is used to check packet lengths */
	if (hTrans == IDLE)
	  NumberBeats <= 0;
	else if ( hTrans == NONSEQ)
	  NumberBeats <= 1;
	else if ( (hTrans == SEQ) && (hReady == 1) )
	  NumberBeats <= NumberBeats + 1;
	else
	  NumberBeats <= NumberBeats;
	/* determine the predicted address */
	if ((hTrans != BUSY) && (hReady != 0))
	  prev_hAddrPlusSize <= hAddr + AddrIncrement;
  end
  /**************************** PSL ASSERTIONS **************************/
  /* define default clock that will be used to sample all subsequent asserts */
  /* NOTE: it is important that data does not change on the same edge as clock
   because, like HDL, assertions can be sensitive to races */
  // psl default clock = (posedge hClk);
  
  /* Behavior: If reset is active then Transfer Type is IDLE and Resp is OKAY
   (default clock) */
  // psl property IdleAndOkayDuringReset = always(
  // {(hResetn == 0);(hResetn == 1)} |->
  // {(hTrans == IDLE) && (hResp == OKAY)});
  
  /* Behavior: If the transfer type is NONSEQ and Burst Mode is SINGLE,
   then on the next clock cycle, the transfer type is not SEQ and not BUSY
   (default clock) */
  // psl property BusyAndSeqNeverFollowNonseqOnSingleTransfer = always (
  // (hTrans == NONSEQ) && (hBurst == SINGLE ) ->
  // next ((hTrans != SEQ) && (hTrans != BUSY)) );
  
  /* Behavior: If the transfer type is BUSY and slave is ready,
   then on the next clock cycle, the transfer type must not be IDLE and
   must not be NONSEQ unless grant is 0 and ready is 1 (default clock) */
  // psl property NeverGoFromBusyToIdleOrNonseqUnlessNoGrant = always (
  // ((hTrans == BUSY) && (hReady == 1)) ->
  // next ((hTrans != IDLE) && (hTrans != NONSEQ))
  // abort ((hGnt == 0) && (hReady == 1)) || (hResetn == 0));
  
  /* Behavior: if the transfer type is IDLE, then on the next clock
   cycle, the transfer type must be either IDLE or NONSEQ (default clock)*/
  // psl property AlwaysGoToNonseqFromIdle = always (
  // (hTrans == IDLE) ->
  // next ((hTrans == IDLE) || (hTrans == NONSEQ)) );
  
  /* Behavior: if the transfer type is BUSY then on the corresponding
   data transfer , the slave must provide a zero wait state OKAY response
   (default clock) */
  // psl property ResponseToBusyMustBeZeroWaitOKAY = always (
  // ((hTrans == BUSY) && (hReady == 1)) ->
  // next ((hResp == OKAY) && (hReady == 1)) );
  
  /* Behavior: if the transfer type is IDLE then on the corresponding data
   transfer, the slave must provide a zero wait state OKAY response
   (default clock) */
  // psl property ResponseToIdleMustBeZeroWaitOKAY = always (
  // ((hTrans == IDLE) && (hReady == 1)) ->
  // next ((hResp == OKAY) && (hReady == 1)) );
  
  /* Behavior: if the previous response was OKAY, and the current response
   is not OKAY, then Ready must be 0 during the second not OKAY response
   (default clock) */
  // psl property ReadyMustBeZeroDuringFirstNotOkayResponse = always (
  // {(hResp == OKAY);(hResp != OKAY)} |->
  // {hReady == 0}) ;
  
  /* Behavior: if the response is SPLIT or RETRY, the master must drive IDLE
   (default clock) */
  // psl property FirstNotOkayResponseCausesNextIdle = always (
  // ((hResp == SPLIT) || (hResp == RETRY)) && (hReady == 0) ->
  // next (hTrans == IDLE));
  
  /* Behavior: if the transfer type is SEQ or BUSY, then the controls of the
   corresponding data transfer must be the same as their previous control
   (default clock) NOTE: the equivalence operator is used with hProt
   because this is an optional signal and may be an "X" value if not
   connected */
  // psl property ControlMustBeConstantDuringABurst = always (
  // (hTrans == SEQ) || (hTrans == BUSY) ->
  // ((hSize == prev_hSize) && (hProt === prev_hProt) &&
  // (hWrite == prev_hWrite)) );
  
  /* Behavior: if Ready is 0 and the response is OK or ERROR, the address
   and control, and data from the master are held constant (default clock)
   NOTE: the equivalence operator is used with hProt because this is an
   optional signal and may be an "X" value if not connected */
  // psl property ControlMustBeConstantWhenSlaveNotReady = always (
  // (hReady == 0) && ((hResp == OKAY) || (hResp == ERROR)) -> next
  // (hTrans == prev_hTrans) && (hSize == prev_hSize) &&
  // (hProt === prev_hProt) && (hWrite == prev_hWrite) &&
  // (hAddr == prev_hAddr) && (hBurst == prev_hBurst) 
  // abort (hResetn == 0));
  
  /* Behavior: if ready is 0 and the previous transfer type was SEQ or NONSEQ,
   then on the next cycle the write data is the same as the previous write data
   (default clock) (This assertion is designed to fail!!!) */
  // psl property WriteDataMustBeConstantWhenSlaveNotReadyAndDataBeingTransferred = always (
  // (hReady == 0) && (hWrite == 1) &&
  // ((prev_hTrans == SEQ) || (prev_hTrans == NONSEQ)) -> next
  // (hWData == prev_hWData) 
  // abort (hResetn == 0));
  
  /* Behavior: if transfer type is BUSY then address does not change
   (default clock) */
  // psl property AddrHeldWhenMasterBusy = always (
  // (hTrans == BUSY) -> next (hAddr == prev_hAddr) );
  
  /* Behavior: if transfer is in progress, then 1K boundary is not exceeded
   (default clock) */
  // psl property PageAddressNeverExceeds1kBoundary = always (
  // ((hTrans == SEQ) || (hTrans == BUSY)) ->
  // (hAddr[31:10] == prev_hAddr[31:10]) );
  
  /* Behavior: If the burst type is not INCR and the transfer type is not
   IDLE , then the NumberBeats for the Burst Mode is not exceeded
   (default clock) */
  // psl property BurstIsNotTooLong = always (
  // (hBurst != INCR) && (hTrans != IDLE) ->
  // ((hBurst == SINGLE) && (NumberBeats <= 1)) ||
  // ((hBurst == WRAP4) && (NumberBeats <= 4)) ||
  // ((hBurst == WRAP8) && (NumberBeats <= 8)) ||
  // ((hBurst == WRAP16) && (NumberBeats <= 16)) ||
  // ((hBurst == INCR4) && (NumberBeats <= 4)) ||
  // ((hBurst == INCR8) && (NumberBeats <= 8)) ||
  // ((hBurst == INCR16) && (NumberBeats <= 16)) );
  
  /* Behavior: For all but INCR Burst mode, if the end of the packet is being
   transferred as indicated by a transition from SEQ to IDLE when Resp is ok
   then the NumberBeats for the Burst Mode is the max number unless grant is 0
   (default clock) */
  // psl property BurstIsNotTooShort = always (
  // {((hBurst != INCR) && (hResp == OKAY) && (hTrans == SEQ));
  // (hTrans == IDLE) } |->
  // {((prev_hBurst == SINGLE) && (NumberBeats == 1)) ||
  // ((prev_hBurst == WRAP4) && (NumberBeats == 4)) ||
  // ((prev_hBurst == WRAP8) && (NumberBeats == 8)) ||
  // ((prev_hBurst == WRAP16) && (NumberBeats == 16)) ||
  // ((prev_hBurst == INCR4) && (NumberBeats == 4)) ||
  // ((prev_hBurst == INCR8) && (NumberBeats == 8)) ||
  // ((prev_hBurst == INCR16) && (NumberBeats == 16))}
  // abort (hGnt == 0)) ;
  
  /* Behavior: if the transfer type is IDLE , then on the next clock cycle,
   the transfer type is not SEQ and the transfer type is not BUSY
   (default clock) */
  // psl property NeverGoFromIdleToSeqOrBusy = always (
  // (hTrans == IDLE) ->
  // next ((hTrans != SEQ) && (hTrans != BUSY)) );
  
  /* Behavior: RETRY is always asserted for two cycles unless reset is asserted
   (default clock)*/
  // psl property RetryResponseMustPersistTwoCycles = always (
  // {(hResp != RETRY);(hResp == RETRY)} |=>
  // {(hResp == RETRY)}
  // abort (hResetn == 0));
  
  /* Behavior: SPLIT is always asserted for two cycles unless reset is asserted
   (default clock) */
  // psl property SplitResponseMustPersistTwoCycles = always (
  // {(hResp != SPLIT);(hResp == SPLIT)} |=>
  // {(hResp == SPLIT)}
  // abort (hResetn == 0));
  
  /* Behavior: ERROR is always asserted for two cycles unless reset is asserted
   (default clock) */
  // psl property ErrorResponseMustPersistTwoCycles = always (
  // {(hResp != ERROR);(hResp == ERROR)} |=>
  // {(hResp == ERROR)}
  // abort (hResetn == 0));
  /******* incorrectAddressCalculation for 3 example WRAP modes: This calculates
   whether the next address is correct for all wrap modes*******/
  
  /* Behavior: if Burst mode is WRAP4 and Size is 8 bits, then if
   transfer mode is SEQ or BUSY, then if previous transfer mode is not
   busy and previous Ready is not zero, then check that the address is as predicted */
  // psl property CorrectAddressDuringPageSize4BurstWrap = always (
  // ((hBurst== WRAP4) && (hSize == bits8)) ->
  // ((hTrans == SEQ || hTrans == BUSY)) ->
  // ((prev_hTrans != BUSY) && (prev_hReady != 0)) ->
  // ((hAddr[31:2] == prev_hAddr[31:2]) &&
  // (hAddr[1:0] == prev_hAddrPlusSize[1:0])) );
  //
  // psl property CorrectAddressDuringPageSize8BurstWrap = always (
  // ((hBurst== WRAP4) && (hSize == bits16)) ||
  // ((hBurst== WRAP8) && (hSize == bits8)) ->
  // ((hTrans == SEQ) || (hTrans == BUSY)) ->
  // ((prev_hTrans != BUSY) && (prev_hReady != 0)) ->
  // ((hAddr[31:3] == prev_hAddr[31:3]) &&
  // (hAddr[2:0] == prev_hAddrPlusSize[2:0])) );
  //
  // psl property CorrectAddressDuringPageSize2048BurstWrap = always (
  // ((hBurst== WRAP16) && (hSize == bits1024)) ->
  // ((hTrans == SEQ) || (hTrans == BUSY)) ->
  // ((prev_hTrans != BUSY) && (prev_hReady != 0)) ->
  // ((hAddr[31:11] == prev_hAddr[31:11]) &&
  // (hAddr[10:0] == prev_hAddrPlusSize[10:0])) );
  
  /* Behavior: the address must increment by size during all INCR beats */
  // psl property AddressIncBySizeDuringAllBurstIncrBeats = always (
  // ((hBurst == INCR16) || (hBurst == INCR8) ||
  // (hBurst == INCR4)) ->
  // ((hTrans == SEQ) || (hTrans == BUSY)) ->
  // ((prev_hTrans != BUSY) && (hReady != 0)) ->
  // (hAddr == prev_hAddrPlusSize) );
  
  /* Behavior: always address must be aligned to the transfer size */
  // psl property AddressNotAlignedToTransferSize = always (
  // ((hSize == bits8) ||
  // ((hSize == bits16) && (hAddr[0] == 1'b0)) ||
  // ((hSize == bits32) && (hAddr[1:0] == 2'b00)) ||
  // ((hSize == bits64) && (hAddr[2:0] == 3'b000)) ||
  // ((hSize == bits128) && (hAddr[3:0] == 4'b0000)) ||
  // ((hSize == bits256) && (hAddr[4:0] == 5'b00000)) ||
  // ((hSize == bits512) && (hAddr[5:0] == 6'b000000)) ||
  // ((hSize == bits1024) && (hAddr[6:0] == 7'b0000000)) ) );
  
  /* Behavior: always the maximum number of wait states is 16 */
  // psl property NeverMoreThan16WaitStates = always (
  // {(hReady == 1);(hReady == 0)} |=>
  // {{(hReady == 0)[*0..15]};{hReady == 1}}
  // abort (hResp != OKAY));
endmodule
