//-------------------------------------------------------------------------------
// Title      :  AHB Interface PSL Sugar Assertions 
// Project    :  Using PSL/Sugar 
//-------------------------------------------------------------------------------
// File       : mem_slave.v
// Author     : ben cohen  <vhdlcohen@aol.com,
// co-author  :
// co-author  :   
// Company    : vhdlcohen training 
// Created    : 2003-06-20
// Last update: 2003-06-20
// Platform   : 
//-------------------------------------------------------------------------------
// Description: AHB slave interface to idt71v433s10 device.
//            : The idt71v433s10 is external to this module. 
//            : This is a synthsizable design.  
//-------------------------------------------------------------------------------
// Copyright (c) 2003 
//-------------------------------------------------------------------------------
// Revisions  :
// Date        Version  Author  Description
// 2003-06-20  1.0      ben2	Created
// 2003-07-18  2.0      ben2    Several modifications
//                              Changed hwrite Transfer direction to:
//                              HIGH =>  write; LOW => read
//                              ** result of code review
//-------------------------------------------------------------------------------


`timescale 1 ns / 100 ps
module mem_slave (hrdata, //AMBA ahb interface
				  hready,
				  hresp,
				  hwdata,
				  haddr,
				  htrans,
				  hwrite,
				  hsize,
				  hburst,
				  hprot,
				  hsel_0n,
				  hresetn, 
				  hclk,
				  wdata_enb,  //  idt71v433s10
				  idt_data,
				  idt_addr,
				  oe_,
				  ce0_,
				  ce1_,
				  cs0,
				  cs1_,
				  lbo_,
				  gw_,
				  bwe_,
				  bw4_,
				  bw3_,
				  bw2_,
				  bw1_,
				  adsp_,
				  adsc_,
				  adv_);
`include "ahb_param.v"
  
  // amba ahb interface 
  output [31:0] hrdata;   // Slave Read data bus
  output [hresp_width -1:0] hresp;    // Transfer response: OKAY, ERROR, RETRY and SPLIT.
  output 					hready;   // Slave Transfer done, HIGH => finished; LOW => extend transfer.
  input [31:0] 				hwdata;   // master Write data bus  
  input [31:0] 				haddr;    // Address bus, Master 
  input [htrans_width -1:0] htrans;   // Master Transfer type,  NONSEQUENTIAL, SEQUENTIAL, IDLE or BUSY.
  input 					hwrite;   // Transfer direction, HIGH =>  write; LOW => read 
  input [hsize_width -1:0] 	hsize;    // Transfer size, byte (8-bit), halfword (16-bit) or word (32-bit).
  input [hburst_width -1:0] hburst;   // Burst type, Four, eight and sixteen beat, incrementing or wrapping.
  input [3:0] 				hprot;    // Protection control
  input 					hsel_0n;  // Slave selec
  input 					hresetn;  // Reset  active LOW 
  input 					hclk;      // Bus clock,  rising edge of HCLK
  
  //  idt71v433s10
  output 					wdata_enb; // 0-> write data to mem,1-> read data from mem
  output [14:0] 			idt_addr;   // idt71v433s10 address 
  output 					oe_;        // idt71v433s10 output enable, active low  
  output 					ce0_;       // #Zero chip idt71v433s10 chip enable, active low
  output 					ce1_;       // #One chip idt71v433s10 chip enable, active low
  output 					cs0;        // idt71v433s10 chip select, active high
  output 					cs1_;       // idt71v433s10 chip select, active low
  output 					lbo_;       // idt71v433s10 linear burst, low==linear count 
  output 					gw_;        // idt71v433s10 global write 
  output 					bwe_;       // idt71v433s10 byte write enable 
  output 					bw4_;       // idt71v433s10 byte write 
  output 					bw3_;
  output 					bw2_;
  output 					bw1_;
  output 					adsp_;      // idt71v433s10 address status process, 1'b0 -> load address 
  output 					adsc_;      // idt71v433s10 address status cache, tie to 1'b1
  output 					adv_ ;      // idt71v433s10 burst address advance 
  input [31:0] 				idt_data;   // idt71v433s10 bidirectional data
  
  
  wire [31:0] 				hrdata;   
  reg [hresp_width -1:0] 	hresp;    
  reg 						hready;
  reg 						wdata_enb;
  wire [14:0] 				idt_addr;  
  reg 						oe_;       
  reg 						ce0_;      
  reg 						ce1_;      
  reg 						cs0;       
  reg 						cs1_;      
  reg 						lbo_;      
  reg 						gw_;       
  reg 						bwe_;      
  reg 						bw4_;      
  reg 						bw3_;
  reg 						bw2_;
  reg 						bw1_;
  reg 						adsp_;     
  reg 						adsc_;     
  reg 						adv_ ; 
  
  
  // ********** Default Clock ********** 
  // psl default clock = (posedge hclk);
  //// psl default clock = (negedge hclk);
  // 
  // IDT idt71v433s10 local parameters and support reg & wires
  
  parameter 				number_idtmem = 0; // 0 -> full board, 1 -> 1 mem unit 
  parameter 				addr_msb = 14;
  parameter 				mem_top = 32767;
  
  reg 						ce_;   // idt71v433s10 chip enable, active low
  //wire 				 hsel_0n; // idt71v433s10 #0 select 
  // wire   hsel_1n; // idt71v433s10 #1 select -- for growth potential 
  
  integer 					burst_advcount;  // keeps track of #times adv_ is 1'b0
  wire 						done_xfrn;  // if 1'b0 -> Slave is done with the current transfer.
  // This signal will be inactive (1'b1) when initial pipeline
  // is filling, and transaction is not done.  It will be active
  // (1'b0) when pipe is filled.
  // Slave machine registers and buffers for address and data
  // 7/14/03 change to full range to accomate multiple idt devices 
  // 7/14/03 Adding 2nd IDT memory
  // 7/09/  New // not for synthsis
  // for synthesis 
  //  wire [31:0] 		 wbuffdata; // Buffered write hwdata
  
  
  reg [31:0] 				addr;  // mem address loaded from haddr
  reg [31:0] 				addr1, addr2, addr3;  // pipelined addr
  reg [31:0] 				out_data; // registered slave data
  reg [32767:0] 			datawritten; // for verification, 1'b1 -> written 
  reg 						next_fsm_error; // input to fsm to guarentee a 2 cycle for hresp=ERROR
  reg 						fsm_error; // register for fsm to handle the hresp when ERROR
  
  
`ifndef SYNTHESIS   
  reg [1:0] 				htrans_r1; // Master Transfer type,  pipe 1
  reg [1:0] 				htrans_r2; // Master Transfer type,  pipe 2
  reg [1:0] 				htrans_r3; // Master Transfer type,  pipe 3
  reg [1:0] 				htrans_r4; // pipe 4
  reg [1:0] 				htrans_r5; // pipe 5
`endif
  
  
  // ***** Design assertion to ensure proper handling at RTL ****
  // 7/05/03 << That assertion is not correctly stated  
  // // done_xfrn==1'b0, but hready is 1'b1
  // // psl property DoneXfr_Hready =
  // //    never (done_xfrn==1'b0 -> hready);
  // Should be (7/0503)
  // done_xfrn==1'b1, but hready is 1'b1
   // 72103  This is an incorrect assumption .. delete 
  //// psl property DoneXfr_Hready =      
  ////     never (done_xfrn==1'b1 -> hready==1'b1);
  // A write fsm machine is envisioned.
  // At this time, the following states are envisioned
  // 7/10/03  Error in size <<<
  // 7/18/03 change width for fsm_write and next_fsm_write to 2
  //  result of code review
  
  
  reg [fsm_write_width-1:0] 				fsm_write; // can be in IDLE, BUSY, WRITE state
  reg [fsm_write_width-1:0] 				next_fsm_write;  // combinational to fsm_write 
  reg [fsm_read_width-1:0] 				next_fsm_read;  // combinational to fsm_read
  reg [fsm_read_width-1:0] 				fsm_read; // can be in IDLE, BUSY, READ state
  
  




   // ******************** MEM_WRITES ********************
   //| # Pattern name # MEM_WRITES
   //| # Problem #  The application does not call for, or rely on burst writes since the 
   //| addresses are always available in the ahb interface by the master.  
   //| To keep the design simple, NON_BURST_WRITE mode is recommended 
   //| by the master (i.e., SINGLE).  But other modes will have no effect on
   //| the operation.  NO check is enformed by hardware for any
   //| protocol violation. In addition, only 32-bit WORD size writes, versus byte, will be
   //| supported in this design.
   //| Busy cycles will be honored. 
   //| # Motivation # Simplicity and conformity.
   //| # Context # Only global word writes will be honored.
   //| # Solution # For a WRITE the following is required:
   //|    @cycle x   adsp_ = 1'b0  // load address
   //|               adsc_ = 1'b1
   //|               adv_  = 1'b1  // no advance 
   //|               lbo_  = 1'b0  // linear burst mode sequence 
   //|               ce_   = 1'b0  // enable chip 
   //|               cs0   = 1'b1  // chip select
   //|               cs1_  = 1'b0  // chip select
   //|               gw_   = 1'b0  // global write 
   //|               {bwe_, bw4_, bw3_, bw2_, bw1_} == 5'b11111 }; // could be don't care,
   //|                         but will enforce for consistency 
   //| # Considerations #  Powerdown ZZ signal is not modeled, and should be 
   //|                in the deasserted dated, i.e. ZZ = 1'b0
   // ********** Check hResp and hready in response to a NONSEQ write cycle **********
   // ** New address cycle when htrans==NONSEQ 
   // psl sequence qNonSeqWriteTransfer = 
   //                {done_xfrn==1'b0 && htrans==NONSEQ && 
   //                 hburst==SINGLE && hwrite==1'b1 && hsel_0n==1'b0 };
   // 
   // psl property HrespOKHready1ForNonSeqWriteTransfer =
   //   always({qNonSeqWriteTransfer} |-> {hresp==OKAY && hready==1'b1});
   // 
   // psl assert HrespOKHready1ForNonSeqWriteTransfer;
   // 
   // ********** Check that writes are in SINGLE bursts **********
   // ** This is a check to ensure design assumption are maintained during verification 
   // psl sequence qWriteTransfer =
   //      {done_xfrn==1'b0 &&  hwrite==1'b1 && hsel_0n==1'b0 };
   //
   // psl property NeverWriteInOtherThanSingleBurst =
   //   always ({qWriteTransfer} |-> {hburst == SINGLE});
   //  BAD never ({qWriteTransfer} |-> {(hburst != SINGLE)});  
   // 
   // psl assume NeverWriteInOtherThanSingleBurst;
   // 
   // ********** Check states of WRITE FSM  **********
   // ** The WRITE FSM states: FSMW_IDLE, FSMW_BUSY, FSMW_WRITE 
   // ** Check of proper state transition, in response to a new WRITE request
   // psl property FsmWriteAfterNonSeqCycle = 
   //   always({qNonSeqWriteTransfer} |=> {fsm_write==FSMW_WRITE});
   //
   // psl assert FsmWriteAfterNonSeqCycle;
   // 
   // ** Check state and hready when transfer is write in BUSY 
   // psl property FsmWriteState2Busy = 
   //       always({fsm_write==FSMW_WRITE && htrans==BUSY} 
   //                    |=> {hready==OKAY && fsm_write==FSMW_BUSY});
   //
   // psl assert FsmWriteState2Busy;
   // 
   //  ** FSMW busy to Write
   // psl sequence qWriteTransferInBusy2Write =
   //      {done_xfrn==1'b0 &&  hwrite==1'b1 && hsel_0n==1'b0 
   //           && htrans==NONSEQ && fsm_write==FSMW_BUSY};
   //
   // psl property FsmWriteBusy2Write=
   //   always({qWriteTransferInBusy2Write} |=> {fsm_write==FSMW_WRITE});
   //
   // ** Check FSMW to IDLE mode
   // ** in IDLE or read mode 
   // psl property FsmWriteState2Idle = 
   //     always({htrans==IDLE || 
   //             (done_xfrn==1'b0 &&  hwrite==1'b0)} |=> {fsm_write==FSMW_IDLE});
   //
   // psl assert FsmWriteState2Idle;
   //
   // ** check FSM to 
   //  ** Ensure for verification that htrans is never in SEQ for WRITE
   // psl sequence qSeqWriteTransfer = 
   //       {done_xfrn==1'b0 && htrans==SEQ && hwrite==1'b1 && hsel_0n==1'b0 };
   // 
   // psl property SeqWriteTransfer= 
   //      never (qSeqWriteTransfer);
   //
   // psl assert SeqWriteTransfer;
   // 
   // ********** Check hready and proper memory controls for Write to memory **********
   // ** Check load of memory address, and write of data in cycles following the SEQ,
   // ** provided htrans is not BUSY 
   // psl sequence qWriteToMem = 
   //     {done_xfrn==1'b0 && fsm_write==FSMW_WRITE && htrans!= BUSY};
   //
   // ** Write control signals to idt71v433s10 device
   //   7/05/03  failed in 1st run ??
    // caused by incorrect assumption in TB in wdata_enb
    // other error in ";" always ({qWriteToMem} |-> { hready==1'b1; 
   // TB is correct, 0-> write, 1-> read , fixed this code
   //  7/07 error in handling od IDT.
   // must change for writes: from adsc_ = 1'b1 TO adsc_ = 1'b0
   //                         from adsp_ = 1'b0 TO adsp_ = 1'b1
   // psl property WriteToIdtMem =
   //   always ({qWriteToMem} |-> { hready==1'b1 && 
   //        adsp_ == 1'b1 && adsc_ == 1'b0 && adv_ == 1'b1 &&
   //        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   //        cs1_  == 1'b0 && gw_   == 1'b0 && bwe_ == 1'b1 && 
   //        ({bwe_, bw4_, bw3_, bw2_, bw1_} == 5'b11111) && 
   //         wdata_enb==1'b0});
   //
   //  psl assert WriteToIdtMem;
   //
   // ** Check wdata_enb is in WRITE direction, only when in qWriteToMem sequence
   // psl property WriteDataEnb = 
   //     always ({qWriteToMem} |-> {wdata_enb==1'b0});
   // BAD    never ({qWriteToMem} |-> {wdata_enb==1'b1});  
   //
   // psl assert WriteDataEnb;
   // 
   //  ** check for no writes to mem when htrans is BUSY or fsm_write is
   //  ** in states other thatn FSMW_WRITE
   // psl sequence qNotWriteSeq = 
   //     {done_xfrn==1'b0 && htrans==BUSY && 
   //       (fsm_write==FSMW_BUSY | fsm_write==FSMW_IDLE | fsm_read==FSMR_IDLE)};
   //
   // psl property NoWriteToIdtMem =
   //   always ({qNotWriteSeq} 
   //        |-> { adsp_ == 1'b1 && adsc_ == 1'b1 && adv_ == 1'b1 &&
   //              lbo_  == 1'b0 && ce_   == 1'b1 && cs0   == 1'b1 &&
   //              cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && 
   //              ({bwe_, bw4_, bw3_, bw2_, bw1_} == 5'b11111) && 
   //                wdata_enb==1'b1});
   //  
   // psl assert NoWriteToIdtMem;  
   //   ------------------------------------------------------------
   // ******************** MEM_READS********************
   //| # Pattern name #  MEM_READS
   //| # Problem #   The architecture with the IDT device provides the first 
   //| data output in the third cycle after the master asserts the address 
   //| during the NONSEQ htrans cycle. In addition, for burst, the IDT device 
   //| allows for the advance mode to increment the address in a 
   //| WRAP4 burst mode.  If the htrans==BUSY, then the IDT device needs to be put on hold.
   //| During the memory pipeline delay cycles, the slave must set the hready to 1'b0.
   //| Only HBurst of SINGLE and WRAP4 will be accepted because of limitation of memory device. 
   //| # Motivation #  Efficiency.  Restriction for WRAP4 is set by memory device selection.
   //|                 Other motivation is design simplicity of slave controller. 
   //| # Context #  Reads in SINGLE or WRAP4 only.  
   //| # Solution #
   //|  Reads in SINGLE or WRAP4 ONLY
   //|  1. The 2 cycles following htrans==NONSEQ, the hready shall be 1'b0.
   //|  2. If hburst == SINGLE then adv_ remains 1'b1 (no advance)
   //|     FSM cycles:  FSMR_IDLE |=>  FSMR_READ_P1 |=> FSMR_READ_P2 |=> FSMR_READ_P3
   //|  3. If hburst == WRAP4 then fsm_read will transition
   //|      FSMR_IDLE |=>  FSMR_READ_P1 |=> FSMR_READ_P2 |=> FSMR_READ_P3
   //|                |=>  FSMR_READ_P4 |=> FSMR_READ_P5 |=> FSMR_READ_P6   
   //|      FSM many remain in FSMR_READ_P4, FSMR_READ_P5, or FSMR_READ_P6
   //|      if htrans==BUSY.
   //|
   //|     if htrans==SEQ |-> 
   //|              adv_ = 1'b0 in FSMR_READ_P2, FSMR_READ_P3, or FSMR_READ_P4 
   //|
   //| # Considerations #   Powerdown ZZ signal is not modeled, and should be 
   //|                in the deasserted dated, i.e. ZZ = 1'b0
   // 
   // ********** Check all reads to be hburst SINGLE or WRAP4 ONLY **********
   // psl sequence qReadTransfer =
   //      {done_xfrn==1'b0 &&  hwrite==1'b0 && hsel_0n==1'b0 };
   //
   // psl property ReadInSingleOrWrap4 =
   //   always ({qReadTransfer} |-> {hburst==SINGLE | hburst== WRAP4});
   // 
   // psl assume ReadInSingleOrWrap4; 
   // 
   // ********** Check hResp to NONSEQ READ cycle **********
   // 7/15/03 need to differentiate between 1st xfr, and subsequent
   //  cycles of the read.  Add the done_xfrn
   // psl sequence qNonSeqReadTransfer = 
   //           {htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0 && done_xfrn==1'b0};
   // 7/15/03 Correct error in design that was defined in Excel timing 
   // psl property HrespOKHready1InNonSeqReadTransfer =
   //   always({qNonSeqReadTransfer} |-> 
   //          {hresp==OKAY && hready==1'b1;
   //           hresp==OKAY && hready==1'b0; 
   //           hresp==OKAY && hready==1'b0;
   //           hresp==OKAY && hready==1'b1}) @ (negedge hclk);
   // 
   // psl assert HrespOKHready1InNonSeqReadTransfer;
   // 
   // ********** Check READ FSM and hready for read cycles **********
   // 7/15/03 insure single 1st read sequence
   // psl sequence qNonSeqReadTransferSingle = 
   //        {hburst==SINGLE && htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0 && done_xfrn==1'b0};
   // 7/15/03 insure single 1st read sequence
   // psl sequence qNonSeqReadTransferWrap4 = 
   //        {hburst==WRAP4 && htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0 && done_xfrn==1'b0};
   //
   // 7/21/03 Change due to error detected in compliance PSL
  //// psl property ControlMustBeConstantWhenSlaveNotReady = always (
  //// (hready == 0) && ((hresp == OKAY) || (hresp == ERROR)) -> next
  //// (htrans == prev_htrans) && (hsize == prev_hsize) &&
  //// (hprot === prev_hprot) && (hwrite == prev_hwrite) &&
  //// (haddr == prev_haddr) && (hburst == prev_hburst) 
  //// abort (hresetn == 0));
  // 721 fail?? 
   // psl property ReadFsmPipelineSingle =
   //     always({qNonSeqReadTransferSingle} |=> 
   //         {fsm_read==FSMR_READ_P1 && hready==1'b0; 
   //          fsm_read==FSMR_READ_P2 && hready==1'b0; 
   //          fsm_read==FSMR_READ_P3 && hready==1'b1; 
   //          (fsm_read==FSMR_IDLE)});
   //was          (fsm_read==FSMR_READ_P1 || fsm_read==FSMR_IDLE)});
   //
   // psl assert ReadFsmPipelineSingle;
   // ----------
   // psl property ReadFsmPipelineWrap4 =
   //     always({qNonSeqReadTransferWrap4} |=>
   //         {fsm_read==FSMR_READ_P1 && hready==1'b0; 
   //          fsm_read==FSMR_READ_P2 && hready==1'b0; 
   //          fsm_read==FSMR_READ_P3 && hready==1'b1;
   //          ((fsm_read==FSMR_READ_P4 && htrans==SEQ && hready==1'b1) || 
   //            fsm_read==FSMR_READ_P3 && htrans==BUSY && hready==1'b1 || 
   //            fsm_read==FSMR_IDLE && htrans==IDLE  && hready==1'b1  || 
   //            fsm_read==FSMR_READ_P1 && 
   //               (htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0) || 
   //            fsm_read==FSMR_IDLE && hwrite==1'b1)});
   //
   // psl assert ReadFsmPipelineWrap4;
   // 
   // psl property FsmReadP4toP5 =
   //    always({fsm_read==FSMR_READ_P4 && htrans==SEQ} |=> {fsm_read==FSMR_READ_P5});  
   //
   // psl assert FsmReadP4toP5;
   // ----------
   // psl property FsmReadP4toP4 =
   //    always({fsm_read==FSMR_READ_P4 && htrans==BUSY} |=> {fsm_read==FSMR_READ_P4});  
   //
   // psl assert FsmReadP4toP4;
   // ----------
   // psl property FsmReadP5toP6 = 
   //    always({fsm_read==FSMR_READ_P5 && htrans==SEQ} |=> {fsm_read==FSMR_READ_P6});
   //
   // psl assert FsmReadP5toP6;
   // ----------
   // psl property FsmReadP5toP5 =
   //    always({fsm_read==FSMR_READ_P5 && htrans==BUSY} |=> {fsm_read==FSMR_READ_P5});
   //
   // psl assert FsmReadP5toP5;
   // ----------
   // psl property FsmReadP6toNext =
   //    always({fsm_read==FSMR_READ_P6} |=> 
   //              {(fsm_read==FSMR_READ_P6 && htrans==BUSY)   || 
   //               (fsm_read==FSMR_READ_P1 && htrans==NONSEQ 
   //                    && hwrite==1'b0 &  hsel_0n==1'b0) || 
   //               (fsm_read==FSMR_IDLE && 
   //                  (htrans==IDLE || hsel_0n==1'b1 || 
   //                   htrans==NONSEQ && hwrite==1'b1) )}) @ (negedge hclk);
   // 
   // psl assert FsmReadP6toNext;
   // ----------
   // 7/11/03 Below assertion is incorrect and should be deleted
   //// psl property NeverReadP6andSEQ =
   ////     never ({fsm_read==FSMR_READ_P6 && htrans==SEQ});
   //// 
   //// psl assert NeverReadP6andSEQ;
   // 
   // ********** Control of idt71v433s10 signals **********
   // psl property MemReadAdvanceControl = 
   //     always({(fsm_read==FSMR_READ_P2 && htrans==SEQ) ||
   //             (fsm_read==FSMR_READ_P3 && htrans==SEQ) ||
   //             (fsm_read==FSMR_READ_P4 && htrans==SEQ)} |-> {adv_==1'b0});
   //
   // psl assert MemReadAdvanceControl; 
   // ----------
   // ** check load of idt71v433s10 during read, P1, P2, P3
   //  7/05/03 Error in 1st run ??
   //  condition for P2 and P3 needs to be added in RTL
   //  7/11/03 It turns out that it's OK to advance idt71v433s10 on single read
   //   This was done in the implementation, but not in the PSL 
   //// psl property IdtReadControlSingle = 
   ////     always({qNonSeqReadTransferSingle} |=>
   ////     {( adsp_ == 1'b1 && adsc_ == 1'b0 && adv_ == 1'b1 &&
   ////        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   ////        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b1); 
   ////      ( adsp_ == 1'b1 && adsc_ == 1'b1 && adv_ == 1'b1 &&
   ////        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   ////        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b0);
   ////      ( adsp_ == 1'b1 && adsc_ == 1'b1 && adv_ == 1'b1 &&
   ////        lbo_  == 1'b0 && ce_   == 1'b1 && cs0   == 1'b1 &&
   ////        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b1)
   ////     });
   // NOTE: this property works only when (htrans!=BUSY) during the read.
   //  This property was changed due to RTL design that took care of don't care conditions
   // 721 <= BUG !! 
   // psl property IdtReadControlSingle = 
   //     always({qNonSeqReadTransferSingle} |=> 
   //     {( adsp_ == 1'b1 && adsc_ == 1'b0 && adv_ == 1'b1 &&
   //        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   //        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b1); 
   //      ( adsp_ == 1'b1 && adsc_ == 1'b1 && (adv_ == 1'b1 || adv_==1'b0) &&
   //        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   //        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b0);
   //      ( adsp_ == 1'b1 && adsc_ == 1'b1 && (adv_ == 1'b1 || adv_==1'b0) &&
   //        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   //        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b0)
   //     });
   //
   // psl assert IdtReadControlSingle;
   // 
   // ** P1, P2
   // psl property IdtReadControlP1 =
   //     always({fsm_read==FSMR_READ_P1} |->
   //     {( adsp_ == 1'b1 && adsc_ == 1'b0 && adv_ == 1'b1 &&
   //        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   //        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b1)
   //     });
   //
   // psl assert IdtReadControlP1;
   // ----------
   // psl property IdtReadControlP2_3_4 =
   //     always({(fsm_read==FSMR_READ_P2 ||
   //              fsm_read==FSMR_READ_P3 ||
   //              fsm_read==FSMR_READ_P4) && htrans==SEQ} |->
   //     {( adsp_ == 1'b1 && adsc_ == 1'b1 && adv_ == 1'b0 &&
   //        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   //        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b0)
   //     });
   //
   // psl assert IdtReadControlP2_3_4; 
   // ----------
   //
   // psl property IdtReadControlP5 =
   //     always({fsm_read==FSMR_READ_P5 && htrans==SEQ} |->
   //     {( adsp_ == 1'b1 && adsc_ == 1'b1 && adv_ == 1'b1 &&
   //        lbo_  == 1'b0 && ce_   == 1'b0 && cs0   == 1'b1 &&
   //        cs1_  == 1'b0 && gw_   == 1'b1 && bwe_ == 1'b1 && oe_==1'b0)
   //     });
   //
   // psl assert IdtReadControlP5;
   // ----------
   // 7/11/03 error in property IdtReadAdvance, the htrans is not needed. 
   // //    never ({(FSMR_READ_P2 || FSMR_READ_P3 || FSMR_READ_P4) 
   // //              htrans==SEQ} |->  {adv_==1'b1}); 
   //
   // psl property IdtReadAdvance =
   //     always ({(fsm_read==FSMR_READ_P2 || 
   //              fsm_read==FSMR_READ_P3 || 
   //              fsm_read==FSMR_READ_P4)} |->  {adv_==1'b0}); 
   // BAD    never ({(fsm_read==FSMR_READ_P2 || 
   //              fsm_read==FSMR_READ_P3 || 
   //              fsm_read==FSMR_READ_P4)} |->  {adv_==1'b1}); 
   //
   // psl assert IdtReadAdvance; 
   //--------------------------------------------------
   //| # Pattern name # ERROR HANDLING  
   //| # Problem #  A slave may have 2 or more idt71v433s10 devices. 
   //|   However, in some configurations, a device may not be installed. 
   //|   It is important that the controller provide an ERROR response 
   //|   if a non-installed memory is addressed
   //|   Design assumes that hburst shall be 
   //|      - SINGLE for writes, and 
   //|      = SINGLE or WRAP4 for reads
   //|   Any other hburst modes should flag and error if device is addressed.
   //| AHB LRM requires 2 cycles of Busy.  That needs to be verified. 
   //| 
   //| # Motivation # Allow growth in the controller design.
   //|    Also, provide feedback to master about an ERROR condition.
   //|
   //| # Context # Installation of a controller with less 
   //|   with less idt71v433s10 devices than what the board can hold.
   //| 
   //| # Solution #
   //| - Detect thru a configuration pin or parameter for the number of idt71v433s10 
   //| devices installed in the board. 
   //| Provide the ERROR indication per ahb specification if an error occurs.
   //| - Ensure when an error occurs:
   //|    (hresp==ERROR  && hready==1'b0) followed by
   //|    (hresp==ERROR  && hready==1'b1) 
   //| # Considerations #  
   //|                
   // **********
   // // 0 -> full board, 1 -> 1 mem unit
   // Error in sequence. ~addr[15]
   //// psl sequence qAccessToNonPrimamryMemory = 
   ////      {(number_idtmem == 1) && done_xfrn==1'b0 && htrans==NONSEQ && hsel_1n==1'b0 };
   // psl sequence qAccessToNonPrimamryMemory = 
   //      {(number_idtmem == 1) && done_xfrn==1'b0 && htrans==NONSEQ && hsel_0n==1'b0 && addr[15]==1'b1 };
   //    
   // psl property MemCountErrorDetect =
   //    always ({qAccessToNonPrimamryMemory} |-> {hready==1'b0; hresp==ERROR && hready==1'b1});   
   // 
   // psl assert MemCountErrorDetect;
   //----------
   // psl sequence qNonSeqWriteTransferNonSingle = 
   //                {done_xfrn==1'b0 && htrans==NONSEQ && 
   //                 hburst!=SINGLE && hwrite==1'b1 && hsel_0n==1'b0 };
   // 
   // psl property HrespErrorForNonSingleWriteTransfer =
   //   always({qNonSeqWriteTransferNonSingle} |-> {hresp==ERROR && hready==1'b1});
   // 
   // psl assert HrespErrorForNonSingleWriteTransfer;
   // ----------
   // psl sequence qNonSeqReadTransferNotSingleNotWrap4 = 
   //           {htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0 && 
   //            !(hburst==SINGLE | hburst==WRAP4)};
   // 
   // psl property NonSeqReadTransferNotSingleNotWrap4 =
   //   always({qNonSeqReadTransferNotSingleNotWrap4} |-> {hresp==ERROR && hready==1'b1});
   // 
   // psl assert NonSeqReadTransferNotSingleNotWrap4; 
   // ----------
   //  psl property HrespToNonseqErrorIsTwoCycles=
   //    always({(hresp==ERROR) && (htrans==NONSEQ)} |-> 
   //             {hready==1'b0; (hready==1'b1) && (hresp==ERROR)});
   //  
   //| # Pattern name # TAGGED TRANSACTION  
   //| # Problem #   Ensure that transactions consisting of multiple-out-of-order 
   //|   write responses match the appropriate read responses. 
   //| 
   //| # Motivation #  Ease of error detection during debug
   //| # Context #  This pattern is useful because of potential memory
   //|    write or read error in single or burst mode, particularly with rollover. 
   //| # Solution # 
   //|    1. A vector shall keep track if data was written into the memory
   //|    2. For 1st idt71v433s10 memory: 
   //|        Data written to memory shall be {1'b0,idt_addr[14:0], 1'b0, idt_addr[14:0]}
   //|       2nd idt71v433s10 memory data=
   //|            {1'b1,idt_addr[14:0], 1'b1, idt_addr[14:0]}
   //|    3. If data is written, then read data[idt_addr] must equal predefined pattern.
   //|    4. Cycle timing indicates that at 
   //|        FSMR_READ_P3, FSMR_READ_P4, FSMR_READ_P5, FSMR_READ_P6 the 
   //|        hwdata is for address x located in register addr.  
   //|        
   //| # Considerations #  For formal verification, ignore this assumption.  
   //|                
   ////  psl property ReadDataToHrData =
   ////    always ((hready==1'b1 && 
   ////            fsm_read==FSMR_READ_P3) -> datawritten[addr3]==1'b1 -> hrdata[7:0]==addr3[7:0]);
   //
   // 7/09/03 new 7/16, fixed range for datawritten address in lookup
   // Need to add the case where datawritten[addr] != 1'b1
   // staticbuild: *W,LRGARR (./mem_slave.v,874|0): The given assignment has a very 
  // large variably indexed node on the lhs, processing the statement may be very time consuming, 
   // you may wish to black-box the module.
   ////  psl property ReadDataWithNothingWritten=
   ////    always ((hready==1'b1 && 
   ////            fsm_read==FSMR_READ_P3) -> datawritten[addr3[addr_msb+2:2]]!==1'b1 -> 1'b0);
   ////  BAD  never ((hready==1'b1 && 
   ////            fsm_read==FSMR_READ_P3) -> datawritten[addr3[addr_msb+2:2]]!==1'b1);
   //
   //// psl assert ReadDataWithNothingWritten; 
   //
   // 7/08/03 Problem reading memory.  Need to add assertions that data read is not X
   //  Use of 'X' in assertions is not supported.
   //// psl property ReadDataIsX =
   ////   always({qNonSeqReadTransfer} |=> {true; idt_data !== 32'bx});
   // BAD  never({qNonSeqReadTransfer} |=> {true; idt_data == 32'bx});
   //// assert ReadDataIsX;


  // addr[15] determines which memory chip is enabled
  // 7/14/03 added 2nd memory chip
  always @ (*) begin
	if (ce_ == 1'b1) begin
	  ce0_ = 1'b1;
	  ce1_ = 1'b1;
	end
	else begin
	  ce0_ = addr[15];  // primary  memory device
	  ce1_ = ~addr[15]; // secondary memory device
	end
  end
  
  
   //
   // ** Done transfer is false during P1 or P2 reads
  // 7/10/03 FIX done_xfrn <<<<<<
  // 72103  Add state for fsm_read==FSMR_READ_P3
  // TO complyto PSL for compliance
  assign 	  done_xfrn= (fsm_read==FSMR_READ_P1 || fsm_read==FSMR_READ_P2 || fsm_read==FSMR_READ_P3) 
                            ? 1'b1 : 1'b0;
  // assign 	  hsel_0n=hsel_sn; // temporary

  // 7/15/03 Compliance PSL demonstrated an error in the interconnection of the
  // address between the mem_slave and the IDT.
  // AMBA LRM, section 3.6 in the note states:
  // All transfers within a burst must be aligned to the adress boundary
  // equal to the size of the size of the transfer. 
  //assign 	  idt_addr = addr[addr_msb:0];
  assign 	  idt_addr = addr[addr_msb+2:2];  

  assign      hrdata = out_data;

  // 7/07/03  Need to tri-state idt_data when not in write mode, and
  //          need to pass hwdata to idt_data when in write mode.
  //  that was left out.  Detected by simulation // no PSL
  //  Need to add a PSL property                            --------
  // assign 	  idt_data =  wdata_enb==1'b0 ? hwdata : 32'hzzzzzzzz;
  // ^^ error since it's done outside in TB 

  // psl property AhbDataToIDT = 
  //   always ({fsm_write==FSMW_WRITE} |-> {idt_data==hwdata}) @ (negedge hclk);
  //  assert AhbDataToIDT;
  
  
   // ********** hresp, hready, and FSM for error  definition **********
   // ** error must be held for 2 cycles 
  always @ (*) begin
	hready=1'b1;
	hresp=OKAY;
	next_fsm_error=1'b0;
	
	// hready response to NONSEQ single burst write 
	if (done_xfrn==1'b0 && htrans==NONSEQ && 
        hburst==SINGLE && hwrite==1'b1 && hsel_0n==1'b0) begin 
	  hready=1'b1;
	  hresp=OKAY; 		  
	end 
	// hready response to BUSY trasnfer type on write
	if (fsm_write==FSMW_WRITE && htrans==BUSY) begin 
	  hready=1'b1;
	  hresp=OKAY; 		  
	end
	if (fsm_read==FSMR_READ_P1 || fsm_read==FSMR_READ_P2) 
	  hready=1'b0;
	  
	if (done_xfrn==1'b0 && htrans==NONSEQ && 
        hburst!=SINGLE && hwrite==1'b1 && hsel_0n==1'b0) begin 
	  hresp=ERROR;
	  hready=1'b0;
	  next_fsm_error=1'b1;
	end
	if (fsm_error)  begin 
	  hresp=ERROR;
	  hready=1'b1;
	  next_fsm_error=1'b0;
	end
  end

  // FSM for writes
  // The WRITE FSM states: FSMW_IDLE, FSMW_BUSY, FSMW_WRITE 
  always @ (*) begin
	// default next state to current state. 
	next_fsm_write = fsm_write; 
	
	case (fsm_write) // write state register
	  // to FSMW_WRITE from FSMW_IDLE
	  FSMW_IDLE: if (done_xfrn==1'b0 && htrans==NONSEQ && 
                     hburst==SINGLE && hwrite==1'b1 && hsel_0n==1'b0) 
        next_fsm_write = FSMW_WRITE;

	  FSMW_WRITE: begin
		// To FSMW_BUSY from FSMW_WRITE
		if (fsm_write==FSMW_WRITE && htrans==BUSY)
          next_fsm_write=FSMW_BUSY;
		// TO FSMW_IDLE when IDLE or a read 
		if (htrans==IDLE || 
			(done_xfrn==1'b0 &&  hwrite==1'b0))  next_fsm_write=FSMW_IDLE;
	  end 
	  
	  FSMW_BUSY: begin 
		// BUSY to idle 
		if (htrans==IDLE)  next_fsm_write=FSMW_IDLE;
		// Busy to Write
		if (done_xfrn==1'b0 &&  
			hwrite==1'b1 && hsel_0n==1'b0 && 
			htrans==NONSEQ && fsm_write==FSMW_BUSY) next_fsm_write=FSMW_WRITE;
	  end 
	endcase
  end


  // FSM for read 
  always @ (*) begin
	next_fsm_read = fsm_read;
	
    case (fsm_read) 
	  FSMR_IDLE: begin
		// 7/10/03  Error in fsm_read.
		// did not complete case for hburst=WRAP4 !!
		// Read in SINGLE burst
        if (hburst==SINGLE && htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0) 
		  next_fsm_read = FSMR_READ_P1;
		// Read in WRAP4 burst 
        if (hburst==WRAP4 && htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0) 
		  next_fsm_read = FSMR_READ_P1;
	  end

	  // Next fsm stage for read
	  FSMR_READ_P1: next_fsm_read = FSMR_READ_P2;
	  
	  // Next fsm stage for read	  
	  FSMR_READ_P2: next_fsm_read = FSMR_READ_P3;
	  
	  //Question:
	  // htrans  NONSEQ  NONSEQ  NONSEQ  NONSEQ
	  // hburst  SINGLE  SINGLE  SINGLE  SINGLE 
	  // hready   1       0        0      1
	  // hwrite   1       1        1      1
	  //                                  ^-- This is not a new read?
	  // 7/21/03 change due to PSL compliance on ready==0 
	  FSMR_READ_P3: begin
		// To P1: NONSEQ new read when current data is  read
		if (htrans==NONSEQ && hwrite==1'b0 && hsel_0n==1'b0) 
		  next_fsm_read = FSMR_IDLE; // next_fsm_read = FSMR_READ_P1;
		
		// To Idle: IDLE or a new write 
		if ((htrans==IDLE) || (hburst==SINGLE && htrans==NONSEQ && hwrite==1'b1 && hsel_0n==1'b0))
		  next_fsm_read = FSMR_IDLE;

		// To P3: Busy, but in Sequential read in WRAP4
		if (hburst==WRAP4 && htrans==BUSY && hwrite==1'b0 && hsel_0n==1'b0)
		  next_fsm_read= FSMR_READ_P3;  // can delete because of default!!!
		
		// To P4: Sequential read in WRAP4
		if (hburst==WRAP4 && htrans==SEQ && hwrite==1'b0 && hsel_0n==1'b0)
		  next_fsm_read= FSMR_READ_P4;	  
	  end 	   
	  
	  FSMR_READ_P4: begin
		// to P4: when busy
		if (htrans==BUSY)   next_fsm_read=FSMR_READ_P4; // can delete because of default!!!
		// to P5: when Sequential read 
		if (htrans==SEQ) next_fsm_read=FSMR_READ_P5;
	  end
	  
	  FSMR_READ_P5: begin 
		// to P5: when busy
		if (htrans==BUSY)   next_fsm_read=FSMR_READ_P5; // can delete because of default!!!
		// to P6: when Sequential read 
		if (htrans==SEQ) next_fsm_read=FSMR_READ_P6;      	   
	  end 
	  
	  
	  FSMR_READ_P6: begin
		// To P6: when busy
		if (htrans==BUSY)   next_fsm_read=FSMR_READ_P6; // can delete because of default!!!
		// To P1: New read 
		if (htrans==NONSEQ && 
            hwrite==1'b0 &  hsel_0n==1'b0) next_fsm_read=FSMR_READ_P1;
		// to Idle: idle, not selected, or a new write
		// PSL bug report showed that stayed ON P6 when htrans==SEQ.
		// Should have gone to IDLE
		// 7/11/03
		//if ((htrans==IDLE || hsel_0n==1'b1 || 
        //     htrans==NONSEQ && hwrite==1'b1)) next_fsm_read=FSMR_IDLE;
		else next_fsm_read=FSMR_IDLE;
	  end
	  
	  //	 FSMR_BUSY: not used 
	endcase
  end

  // Load of FSM registers and addr 
  always @ (posedge hclk or negedge hresetn ) begin
	if (!hresetn) begin 
	  fsm_read <= FSMR_IDLE;
	  fsm_write <= FSMW_IDLE;
	  fsm_error <= 1'b0;
	  addr <= 0;
	end
	else begin
	  fsm_read <= next_fsm_read;
	  fsm_write <= next_fsm_write;
	  fsm_error <= next_fsm_error;
	  // addr <= haddr;
	  out_data <= idt_data;
	  if (htrans != BUSY) begin
		addr <= haddr; // 7/09
		// not for synthesis
        `ifndef SYNTHESIS
		//  assign 	  idt_addr = addr[addr_msb+2:2];  
		addr1 <= haddr;
		addr2 <= addr1;
		addr3 <= addr2;
		`endif 
		// for synthesis
		htrans_r1 <= htrans;
		htrans_r2 <= htrans_r1;
		htrans_r3 <= htrans_r2;
		htrans_r4 <= htrans_r3;
		htrans_r5 <= htrans_r4;
		// not for synthesis
		`ifndef SYNTHESIS
		// 7/11/03 error in formula
		// if (done_xfrn==1'b0 && fsm_write==FSMW_WRITE && htrans!= BUSY)
		// if (fsm_write==FSMW_WRITE && htrans!= BUSY) 		  
		  // datawritten[idt_addr] = 1'b1;
		// for synthesis
		`endif
	  end 
	end
  end	
  
  
  // idt71v433s10 controls
  always @ (*) begin
	// defaults 
	adsp_ = 1'b1; adsc_ = 1'b1; adv_ = 1'b1;
    lbo_  = 1'b0; ce_   = 1'b1; cs0   = 1'b1;
    cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b1;
    {bwe_, bw4_, bw3_, bw2_, bw1_} = 5'b11111; 
    wdata_enb=1'b1; // read mode 
    // write case  
    if (done_xfrn==1'b0 && fsm_write==FSMW_WRITE && htrans!= BUSY) begin
      adsp_ = 1'b1; adsc_ = 1'b0; adv_ = 1'b1;
      lbo_  = 1'b0; ce_   = 1'b0; cs0  = 1'b1;
      cs1_  = 1'b0; gw_   = 1'b0; bwe_ = 1'b1; 
      {bwe_, bw4_, bw3_, bw2_, bw1_} = 5'b11111; 
      wdata_enb=1'b0;
	end
	
	// Failed to initally add these conditions for the read.
	// 7/05/03 
	// from psl property IdtReadControlP1 =
    if (fsm_read==FSMR_READ_P1) begin 
      adsp_ = 1'b1; adsc_ = 1'b0; adv_ = 1'b1;
      lbo_  = 1'b0; ce_   = 1'b0; cs0   = 1'b1;
      cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b1;
	end 
	
    // from psl property IdtReadControlP2_3_4 =
	// ** Error in formula 7/10/03 
    //if ((fsm_read==FSMR_READ_P2 && htrans_r2) ||
    //    (fsm_read==FSMR_READ_P3 && htrans_r3) ||
    //    (fsm_read==FSMR_READ_P4 && htrans_r4==SEQ)) begin 
    //  adsp_ = 1'b1; adsc_ = 1'b1; adv_ = 1'b0;
    //  lbo_  = 1'b0; ce_   = 1'b0; cs0   = 1'b1;
    //  cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b0;
    //end
	// Added this as a result of simulation
	// 7/11/03 
	if ((fsm_read==FSMR_READ_P2 && (htrans!=BUSY)) ||
        (fsm_read==FSMR_READ_P3 && (htrans!=BUSY)) ||
        (fsm_read==FSMR_READ_P4 && (htrans!=BUSY)) ) begin 
      adsp_ = 1'b1; adsc_ = 1'b1; adv_ = 1'b0;
      lbo_  = 1'b0; ce_   = 1'b0; cs0   = 1'b1;
      cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b0;
    end

	// Modified as a result of simulation analys and PSL assertions 
	if (fsm_read==FSMR_READ_P5) begin 
      adsp_ = 1'b1; adsc_ = 1'b1; adv_ = 1'b1;
      lbo_  = 1'b0; ce_   = 1'b0; cs0   = 1'b1;
      cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b0;
    end


	/*
	 * // from psl property IdtReadControlSingle 
    if ((fsm_read==FSMR_READ_P2) && htrans_r2==NONSEQ) begin  // <<bug 
      adsp_ = 1'b1; adsc_ = 1'b1; adv_ = 1'b1;
      lbo_  = 1'b0; ce_   = 1'b0; cs0   = 1'b1;
      cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b0;
    end

		// from psl property IdtReadControlSingle 
    if ((fsm_read==FSMR_READ_P3) && htrans_r3==NONSEQ) begin // bug 
      adsp_ = 1'b1; adsc_ = 1'b1; adv_ = 1'b1;
      lbo_  = 1'b0; ce_   = 1'b1; cs0   = 1'b1;
      cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b1;
    end
	 */
	
	
    // from psl property IdtReadControlP5 =
    //if (fsm_read==FSMR_READ_P5) begin 
    //  adsp_ = 1'b1; adsc_ = 1'b1; adv_ = 1'b1;
    //  lbo_  = 1'b0; ce_   = 1'b0; cs0   = 1'b1;
	//  cs1_  = 1'b0; gw_   = 1'b1; bwe_ = 1'b1; oe_=1'b0;
	//end
  end


endmodule




